<!-- head scripts -->
<script type="text/javascript">
    var spec = "{{ site.baseurl }}/_static/reports/{{page.json}}";
    console.log(spec);
    // var spec = "https://raw.githubusercontent.com/vega/vega/master/docs/examples/bar-chart.vg.json";
    var embedOpt = {"renderer": "svg", "mode": "vega-lite"};

    vegaEmbed('#vis', spec, embedOpt).then(function(result) {
        // Access the Vega view instance (https://vega.github.io/vega/docs/api/view/) as result.view
    }).catch(console.error);
</script>