INDIVIDUAL_OUTPUT = {
    "value": 2.0,
    "mean": None,
    "median": None,
    "min": None,
    "max": None,
    "sd": None,
    "se": None,
    "cv": None,
    "unit": "ng",
    "value_weight": 80.0,
    "unit_weight": "kg",
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}
INDIVIDUAL_NO_BODYWEIGHT_OUTPUT = {
    "value": 2.0,
    "mean": None,
    "median": None,
    "min": None,
    "max": None,
    "sd": None,
    "se": None,
    "cv": None,
    "unit": "ng",
    "value_weight": None,
    "unit_weight": None,
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}
GROUP_OUTPUT = {
    "value": None,
    "mean": 4.5,
    "median": None,
    "min": None,
    "max": None,
    "sd": 4.6,
    "se": None,
    "cv": 0.4,
    "unit": "ng",
    "mean_weight": 80.0,
    "unit_weight": "kg",
    "intervention_value": 0.8,
    "intervention_unit": "ng",
}
