---
---

// Import Main CSS file from theme
@import "main";

.content {
  display:block;
  min-width: 600px;
  overflow-x: scroll;
}

.vega-bind {
  text-align:right;
  position: absolute;
  right:45px;
  top:20px;

	
}

body{
  overflow-x:hidden;
}

select { 
	width: 15em;
	height: 3em 
}
