{% if page.menubar and page.show_sidebar %} 
    {% assign content_width = 'is-4' %} 
{% elsif page.menubar or page.show_sidebar %} 
    {% assign content_width = 'is-8' %} 
{% else %} 
    {% assign content_width = 'is-12' %} 
{% endif %}
<!DOCTYPE html>
<html>
  {% include head.html %}
  <body>
    {% include header.html %}
    {% unless page.hide_hero %}
        {% include hero.html %}
    {% endunless %}
    <section class="section">
        <div class="container">
            <div class="columns">
                {% if page.menubar %}
                <div class="column is-4-desktop is-4-tablet">
                    {% include menubar.html %}
                </div>
                {% endif %}
                <div class="column {{ content_width }}">
                    {{ content }}
                </div>
            </div>
        </div>
    </section>
    <script src="{{ site.baseurl }}/assets/js/app.js" type="text/javascript"></script>
    {%- include footer-scripts.html -%}
  </body>
</html>

