<head>
    <meta charset="utf-8">
    <script src="https://cdn.jsdelivr.net/npm/vega@5"></script>
    <script src="https://cdn.jsdelivr.net/npm/vega-lite@4"></script>
    <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{{ page.title }} - {{ site.title }}</title>
    <link rel="stylesheet" href="{{ site.baseurl }}/assets/css/app.css">
    <script defer src="https://use.fontawesome.com/releases/v5.3.1/js/all.js"></script>
    {% seo %}
    {%- if site.google_analytics -%}
      {%- include google-analytics.html -%}
    {%- endif -%}
    {%- include head-scripts.html -%}
</head>
